#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mrp.h"

POINT dyx[] = {
    /* 1 */
    { 0,-1}, {-1, 0},
    /* 2 */
    { 0,-2}, {-1,-1}, {-1, 1}, {-2, 0},
    /* 3 */
    { 0,-3}, {-1,-2}, {-2,-1}, {-1, 2}, {-2, 1}, {-3, 0},
    /* 4 */
    { 0,-4}, {-1,-3}, {-2,-2}, {-3,-1}, {-1, 3}, {-2, 2}, {-3, 1}, {-4, 0},
    /* 5 */
    { 0,-5}, {-1,-4}, {-2,-3}, {-3,-2}, {-4,-3}, {-1, 4}, {-2, 3}, {-3, 2},
    {-4, 1}, {-5, 0},
    /* 6 */
    { 0,-6}, {-1,-5}, {-2,-4}, {-3,-3}, {-4,-2}, {-5,-1}, {-1, 5}, {-2, 4},
    {-3, 3}, {-4, 2}, {-5, 1}, {-6, 0},
    /* 7 */
    { 0,-7}, {-1,-6}, {-2,-5}, {-3,-4}, {-4,-3}, {-5,-2}, {-6,-1}, {-1, 6},
    {-2, 5}, {-3, 4}, {-4, 3}, {-5, 2}, {-6, 1}, {-7, 0},
};
double sigma[] = {0.85, 1.15, 1.50, 1.90, 2.55, 3.30, 4.25, 5.60,
                  7.15, 9.20,12.05,15.35,19.95,25.85,32.95,44.05};

FILE *fileopen(char *filename, char *mode)
{
    FILE *fp;
    fp = fopen(filename, mode);
    if (fp == NULL) {
        fprintf(stderr, "Can\'t open %s!\n", filename);
        exit(1);
    }
    return (fp);
}

void **alloc_2d_array(int height, int width, int size)
{
    void **mat;
    char *ptr;
    int k;

    mat = malloc(sizeof(void *) * height + height * width * size);
    if (mat == NULL) {
	fprintf(stderr, "Can\'t allocate memory!\n");
	exit(1);
    }
    ptr = (char *)(mat + height);
    for (k = 0; k < height; k++) {
	mat[k] =  ptr;
	ptr += width * size;
    }
    return (mat);
}

IMAGE *alloc_image(int width, int height, int maxval)
{
    IMAGE *img;
    img = (IMAGE *)malloc(sizeof(IMAGE));
    img->width = width;
    img->height = height;
    img->maxval = maxval;
    img->val = (img_t **)alloc_2d_array(img->height, img->width,
                                        sizeof(img_t));
    return (img);
}

int *gen_hufflen(int *hist, int size, int max_len)
{
    int i, j, k, l, *len, *index, *bits, *link;

    len = (int *)malloc(size * sizeof(int));
    index = (int *)malloc(size * sizeof(int));
    bits = (int *)malloc(size * sizeof(int));
    link = (int *)malloc(size * sizeof(int));
    for (i = 0; i < size; i++) {
        len[i] = 0;
        index[i] = i;
        link[i] = -1;
    }
    /* sort in decreasing order of frequency */
    for (i = size -1; i > 0; i--) {
	for (j = 0; j < i; j++) {
	    if (hist[index[j]] < hist[index[j + 1]]) {
                k = index[j + 1];
                index[j + 1] = index[j];
                index[j] = k;
	    }
	}
    }
    for (i = 0; i < size; i++) {
        bits[i] = index[i];	/* reserv a sorted index table */
    }
    for (i = size - 1; i > 0; i--) {
        k = index[i - 1];
        l = index[i];
        hist[k] += hist[l];
        len[k]++;
	while (link[k] >= 0) {
            k = link[k];
            len[k]++;
	}
        link[k] = l;
        len[l]++;
	while (link[l] >= 0) {
            l = link[l];
            len[l]++;
	}
	for (j = i - 1; j > 0; j--) {
	    if (hist[index[j - 1]] < hist[index[j]]) {
                k = index[j];
                index[j] = index[j - 1];
                index[j - 1] = k;
	    } else {
                break;
	    }
	}
    }
    /* limit the maximum code length to max_len */
    for (i = 0; i < size; i++) {
	index[i] = bits[i];	/* restore the index table */
        bits[i] = 0;
    }
    for (i = 0; i < size; i++) {
        bits[len[i]]++;
    }
    for (i = size - 1; i > max_len; i--) {
	while (bits[i] > 0) {
            j = i - 2;
            while(bits[j] == 0) j--;
            bits[i] -= 2;
            bits[i - 1]++;
            bits[j + 1] += 2;
            bits[j]--;
	}
    }
    for (i = k = 0; i < size; i++) {
	for (j = 0; j < bits[i]; j++) {
            len[index[k++]] = i;
	}
    }
    free(link);
    free(bits);
    free(index);
    return (len);
}

void gen_huffcode(VLC *vlc)
{
    int i, j, *idx, *len;
    unsigned int k;

    vlc->index = idx = (int *)malloc(vlc->size * sizeof(int));
    vlc->off = (int *)malloc(vlc->max_len * sizeof(int));
    vlc->code = (unsigned int *)malloc(vlc->size * sizeof(int));
    len = vlc->len;
    /* sort in increasing order of code length */
    for (i = 0; i < vlc->size; i++) {
        idx[i] = i;
    }
    for (i = vlc->size -1; i > 0; i--) {
	for (j = 0; j < i; j++) {
	    if (len[idx[j]] > len[idx[j + 1]]) {
                k = idx[j + 1];
                idx[j + 1] = idx[j];
                idx[j] = k;
	    }
	}
    }
    k = 0;
    for (j = 0; j < vlc->max_len; j++) {
	vlc->off[j] = -1;
    }
    j = len[idx[0]];
    for (i = 0; i < vlc->size; i++) {
	if (j < len[idx[i]]) {
	    k <<= (len[idx[i]] - j);
	    j = len[idx[i]];
	}
	vlc->code[idx[i]] = k++;
	vlc->off[j - 1] = i;
    }
    return;
}

VLC *make_vlc(int *hist, int size, int max_len)
{
    VLC *vlc;

    vlc = (VLC *)malloc(sizeof(VLC));
    vlc->size = size;
    vlc->max_len = max_len;
    vlc->len = gen_hufflen(hist, size, max_len);
    gen_huffcode(vlc);
    return (vlc);
}

VLC **init_vlc(int num_group, int size)
{
    VLC **vlc;
    int i, gr, *hist;
    double a;

    vlc = (VLC **)malloc(num_group * sizeof(VLC *));
    hist = (int *)malloc(size * sizeof(int));
    for (gr = 0; gr < num_group; gr++) {
        /* generate Huffman code */
        a = -1.0 / (8.0 * sigma[gr] * sigma[gr]);
	for (i = 0; i < size; i++) {
            hist[i] = 1E6 * exp(a * (i * i));
	}
        vlc[gr] = make_vlc(hist, size, VLC_MAXLEN);
    }
    free(hist);
    return (vlc);
}

void free_vlc(VLC *vlc)
{
    free(vlc->code);
    free(vlc->off);
    free(vlc->index);
    free(vlc->len);
    free(vlc);
    return;
}

int calc_u(img_t **idx_E, int width, int y, int x)
{
    int rx, ry, k, u;

    u = 0;
    if (y > UPEL_DIST && x > UPEL_DIST && x < width - UPEL_DIST) {
	for (k = 0; k < NUM_UPELS; k++) {
	    ry = y + dyx[k].y;
	    rx = x + dyx[k].x;
	    u += idx_E[ry][rx];
	}
    } else if (y == 0) {
	if (x == 0) {
	    u = (MAX_UPARA >> 1);
	} else {
            ry = 0;
	    for (k =0; k < NUM_UPELS; k++) {
                rx = x + dyx[k].x;
                if (rx < 0) rx = 0;
                else if (rx >= x) rx = x - 1;
                u += idx_E[ry][rx];
	    }
	}
    } else {
	if (x == 0) {
	    for (k = 0; k < NUM_UPELS; k++) {
		ry = y + dyx[k].y;
		if (ry < 0) ry = 0;
		else if (ry >= y) ry = y - 1;
		rx = x + dyx[k].x;
		if (rx < 0) rx = 0;
		u += idx_E[ry][rx];
	    }
	} else {
	    for (k = 0; k < NUM_UPELS; k++) {
		ry = y + dyx[k].y;
		if (ry < 0) ry = 0;
		rx = x + dyx[k].x;
		if (rx < 0) rx = 0;
		else if (rx >= width) rx = width - 1;
		u += idx_E[ry][rx];
	    }
	}
    }
    if (u > MAX_UPARA) u = MAX_UPARA;

    return (u);
}

int e2E(int e, int prd, int flag, int maxval)
{
    int E, th;

    E = (e > 0)? e : -e;
    th = (prd < ((maxval + 1) >> 1))? prd : maxval - prd;
    if (E > th) {
        E += th;
    } else if (flag) {
	E = (e < 0)? (E << 1) - 1 : (E << 1);
    } else {
	E = (e > 0)? (E << 1) - 1 : (E << 1);
    }
    return (E);
}

int E2e(int E, int prd, int flag, int maxval)
{
    int e, th;

    th = (prd < ((maxval + 1) >> 1))? prd : maxval - prd;
    if (E > (th << 1)) {
	e = (prd < ((maxval + 1) >> 1))? E - th : th - E;
    } else if (flag) {
	e = (E & 1)? -((E >> 1) + 1) : (E >> 1);
    } else {
	e = (E & 1)? (E >> 1) + 1 : -(E >> 1);
    }
    return (e);
}
